# -*- coding: utf-8 -*-

import pygame

import context
import gummworld2
import settings


# ----------------------------------------------------------------------------

class _BaseTransition(context.Context):

    def __init__(self, old_context, new_context, pop_old=True):
        self.old_context = old_context
        self.new_context = new_context
        self.pop_old = pop_old

    def suspend(self):
        """Called when another context is pushed on top of this one."""
        self.old_context.suspend()
        self.new_context.suspend()

    def resume(self):
        """Called when another context is popped off the top of this one."""
        self.old_context.resume()
        self.new_context.resume()

    def enter(self):
        """Called when this context is pushed onto the stack."""
        if self.new_context:
            self.new_context.enter()

    def exit(self):
        """Called when this context is popped off the stack."""
        if self.pop_old:
            context.pop() # pops the old context
        if self.new_context:
            context.push(self.new_context, False)

    def think(self, dt):
        """Called once per frame"""
        context.Context.think(self, dt)

    def draw(self, screen):
        """Refresh the screen"""
        pass

# ----------------------------------------------------------------------------

class FadeInTranition(_BaseTransition):
    """fades from old to new context, pops itself and top of stack"""
    def __init__(self, old, new, duration, pop_old=True):
        self.color = (0,0,0)
        self.new_color = (0,0,0)
        if isinstance(old, tuple):
            self.color = old
            old = None
        if isinstance(new, tuple):
            self.new_color = new
            new = None
        super(FadeInTranition, self).__init__(old, new, pop_old)
        self.duration = duration
        self.time = 0.0
        self.alpha = 255.0

    def enter(self):
        super(FadeInTranition, self).enter()
        self.time = 0.0

    def exit(self):
        super(FadeInTranition, self).exit()
        pygame.event.clear()

    def think(self, dt):
        self.time += dt
        if self.time < self.duration:
            self.alpha = int(255.0 - self.time / self.duration * 255.0)
            if self.alpha < 0:
                context.pop()
                return
        else:
            context.pop()
            return
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                if event.key in settings.keys.quit:
                    context.pop()
        
    def draw(self, screen):
        """Refresh the screen"""
        if self.new_context:
            self.new_context.draw(screen)
        else:
            screen.surface.fill(self.new_color)
        size = screen.surface.get_size()
        s = gummworld2.screen.Screen(size, 0, pygame.Surface(size))
        if self.old_context:
            self.old_context.draw(s)
        else:
            s.surface.fill(self.color)
        s.surface.set_alpha(self.alpha)
        screen.surface.blit(s.surface, (0, 0))

# ----------------------------------------------------------------------------
# ----------------------------------------------------------------------------

class FadeOutTranition(_BaseTransition):
    """ fades from a current context to the new one, pop only itself"""
    def __init__(self, old, new, duration, pop_old=True):
        self.color = (0,0,0)
        self.old_color = (0,0,0)
        if isinstance(new, tuple):
            self.color = new
            new = None
        if isinstance(old, tuple):
            self.old_color = old
            old = None
        super(FadeOutTranition, self).__init__(old, new, pop_old)
        self.duration = duration
        self.time = 0.0
        self.alpha = 0.0

    def enter(self):
        if self.new_context:
            super(FadeOutTranition, self).enter()
        self.time = 0.0

    def exit(self):
        # super(FadeOutTranition, self).exit()
        pygame.event.clear()

    def think(self, dt):
        self.time += dt
        if self.time < self.duration:
            self.alpha = int(self.time / self.duration * 255.0)
            if self.alpha > 255.0:
                context.pop()
                return
        else:
            context.pop()
            return
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                if event.key in settings.keys.quit:
                    context.pop()
        
    def draw(self, screen):
        """Refresh the screen"""
        if self.old_context:
            self.old_context.draw(screen)
        else:
            screen.surface.fill(self.old_color)
        size = screen.surface.get_size()
        s = gummworld2.screen.Screen(size, 0, pygame.Surface(size))
        if self.new_context:
            self.new_context.draw(s)
        else:
            s.surface.fill(self.color)
        s.surface.set_alpha(self.alpha)
        screen.surface.blit(s.surface, (0, 0))

# ----------------------------------------------------------------------------
# ----------------------------------------------------------------------------

